//*********************************************************************************************
// edge.cpp
//*********************************************************************************************

#include "StdAfx.h"

#include "permutation.h"

#include "edge.h"

//----------------------------------------------------------------------------------------------

// Check if vertex v is one of the end vertices
bool CEdge::has_endvertex(int v) const
{ 
  return ((_v == v) || (_w == v)); 
};

//----------------------------------------------------------------------------------------------

// Construct edge with specified end vertices
CEdge::CEdge(int v, int w)
{
/*
  // No loops permitted
  assert(v != w);
*/
  // Internally, we enforce the convention v < w.
  _v = min(v, w);
  _w = max(v, w);
}

//----------------------------------------------------------------------------------------------

// Lexicographic order on pairs of end vertices
bool CEdge::operator <(const CEdge& rhs) const
{
  if (_v != rhs._v) 
    {
      return (_v < rhs._v);
    }
  return (_w < rhs._w);
}

//----------------------------------------------------------------------------------------------

// Equality of end vertices
bool CEdge::operator ==(const CEdge& rhs) const
{
  return ((_v == rhs._v) && (_w == rhs._w));
}

//----------------------------------------------------------------------------------------------

// Stream out edge
ostream& operator<<(ostream& os, const CEdge& e)
{
  os << "<" << e._v << ", " << e._w << ">";

  return os;
}

//----------------------------------------------------------------------------------------------

// Total order on edge lists 
// Let E = (e_1, ..., e_s) and F = (f_1, ..., f_t) be ordered lists of edges.
// Precedence is given to edge lists with fewer edges, hence the empty list is least element.
// E < F :<=> s < t             if s != t
//            E < F under lex   otherwise
bool CEdges::operator<(const CEdges& rhs) const
{
  if (size() != rhs.size()) return (size() < rhs.size());
  return (_E < rhs._E);
}

//----------------------------------------------------------------------------------------------

// Equality of edge lists
bool CEdges::operator==(const CEdges& rhs) const
{
  return (_E == rhs._E);
}

//----------------------------------------------------------------------------------------------

// Construct list of s edges (<v_1, w_1>, ..., <v_s, w_s>) from 2*s elementary
// list of end vertices (v_1, w_1, ..., v_s, w_s)
CEdges::CEdges(int two_s, ...)
{
  va_list ap;
  va_start(ap, two_s);

  assert(two_s%2 == 0);
  int s = two_s/2;
  int w, v;

  for (int i=0; i<s; i++)
    {
      v = va_arg(ap, int);
      w = va_arg(ap, int);
      _E.push_back(CEdge(v, w));
    }

  va_end(ap);
}

//----------------------------------------------------------------------------------------------

// Insert edge f before the first edge in the list greater than f under lexicographic ordering 
// on end vertex pairs, provided f is not in the list yet. Otherwise, to avoid duplicates, f
// is discarded and false returned. If the list is ordered initially, it will remain ordered.
bool CEdges::add(const CEdge& f)
{
  vector<CEdge>::iterator e;

  for(e = _E.begin(); e != _E.end(); e++)
    {
      if (f == *e)
        {
          return false;
        }
      else if (f < *e) 
        {
          _E.insert(e, f);
          return true;
        }
    }
  _E.push_back(f);

  return true;
}

//----------------------------------------------------------------------------------------------

// Insert each edge f in F as described above and return the number of successful additions.
int CEdges::add(const CEdges& F)
{
  int inserted = 0;
  vector<CEdge>::const_iterator f;
  
  for(f = F._E.begin(); f != F._E.end(); f++)
    {
      if (add(*f)) inserted++;
    }

  return inserted;
}

//----------------------------------------------------------------------------------------------

// Remove edge f from the list if present, otherwise return false.
bool CEdges::remove(const CEdge& f)
{
  vector<CEdge>::iterator e;

  for(e = _E.begin(); e != _E.end(); e++)
    {
      if (f == *e)
        {
          _E.erase(e);
          return true;
        }
    }

  return false;
}

//----------------------------------------------------------------------------------------------

// Remove each edge f in F as described above and return the number of successful removals.
int CEdges::remove(const CEdges& F)
{
  int removed = 0;
  vector<CEdge>::const_iterator f;

  for (f = F._E.begin(); f != F._E.end(); f++)
    {
      if (remove(*f)) removed++;
    }

  return removed;
}

//----------------------------------------------------------------------------------------------

// Merge two ordered edge lists E1 and E2 into the edge list _E
void CEdges::combine(const CEdges& E1, const CEdges& E2)
{
  merge(E1._E.begin(), E1._E.end(), E2._E.begin(), E2._E.end(), back_inserter(_E));
}

//----------------------------------------------------------------------------------------------

// Determine the smallest vertex v = v_star not incident with any of the edges in E.
// Note: If V = [n] and no isolated vertex exists, this methods returns n+1.
int CEdges::first_isolated_vertex(void)
{
  vector<bool> isolated;
  int n = 0;

  vector<CEdge>::iterator e;
  for(e = _E.begin(); e != _E.end(); e++)
    {
      if (e->_w > n)
        {
          isolated.resize(e->_w);
          for (int i=n; i<e->_w; i++) isolated[i] = true;
          n = e->_w;
        }
      isolated[e->_v-1] = false;
      isolated[e->_w-1] = false;
    }
     
  for (int i=0; i<n; i++)
    {
      if (isolated[i]) return i+1;
    }

  return n+1;
}

//----------------------------------------------------------------------------------------------

// Stream out edge list
ostream& operator<<(ostream& os, const CEdges& E)
{
  int s = E.size();
  
  os << "{";
  for (int e = 0; e < s; e++)
    {
      if (e > 0) os << ", ";
      os << E._E[e];
    }
  os << "}";

  return os;
}

//----------------------------------------------------------------------------------------------

// Count number of end vertices of f contained in the vertex list
int CVertices::find_incidences(const CEdge& f) const
{
  int num_inc = 0;
  int n = _V.size();

  for (int i=0; (i<n) && (num_inc < 2) ; i++)
    {
      if (f.has_endvertex(_V[i]))
        {
          num_inc++;
        }
    }

  return num_inc;
}

//----------------------------------------------------------------------------------------------

// Insert vertex v before the first vertex in the list greater than v, provided v is not in 
// the list yet. Otherwise, to avoid duplicates, v is discarded and false returned. If the list 
// is ordered initially, it will remain ordered.
bool CVertices::add(int w)
{
  vector<int>::iterator v;
  for(v = _V.begin(); v != _V.end(); v++)
    {
      if (w == *v)
        {
          return false;
        }
      else if (w < *v) 
        {
          _V.insert(v, w);
          return true;
        }
    }
  _V.push_back(w);
  return true;
}

//----------------------------------------------------------------------------------------------

// Remove vertex v from the list if present, otherwise return false.
bool CVertices::remove(int v)
{
  vector<int>::iterator w;

  for(w = _V.begin(); w != _V.end(); w++)
    {
      if (v == *w)
        {
          _V.erase(w);
          return true;
        }
    }

  return false;
}

//----------------------------------------------------------------------------------------------

// Merge two ordered vertex lists V1 and V2 into the vertex list _V
void CVertices::combine(const CVertices& V1, const CVertices& V2)
{
  merge(V1._V.begin(), V1._V.end(), V2._V.begin(), V2._V.end(), back_inserter(_V));
}

//----------------------------------------------------------------------------------------------

// Stream out set of vertices
ostream& operator<<(ostream& os, const CVertices& V) 
{
  int n = V.size();
  
  os << "{";
  for (int v = 0; v < n; v++)
    {
      if (v > 0) os << ", ";
      os << V._V[v];
    }
  os << "}";

  return os;
}

//----------------------------------------------------------------------------------------------
