//*********************************************************************************************
// graph.cpp
//*********************************************************************************************

#include "StdAfx.h"

#define EXTDEFS 1
#include "Nauty20/nauty.h"

#include "permutation.h"

#include "graph.h"

// LOW_LAYER indicates from which layer storage-for-performance measures are used
// (set to any number > N = n*(n-1)/2 to deactivate such measures completely)
#define LOW_LAYER 8

// SMALL_CONSTRAINT indicates from what size of constraints X_j values of M^{X_j}(D)
// are deemed sufficiently easy to compute, so that no further recursion is used.
// With n = 7, setting 3 gives approx. 20% memory saving at 80% more computing time.
#define SMALL_CONSTRAINT 3

typedef M_map::value_type M_entry;

//----------------------------------------------------------------------------------------------

// Check inequality M_0 * M_1 <= M_2 * M_3, provide the ratio rhs/lhs and, if required, 
// verify that provided values satisfy M_k = M^{X_k}_{Y_k}(G).

bool CGraph::check_inequality(long *M, const CEdge& e, const CEdge& f, CGraph& G, 
                              double *ratio, bool verbose, bool check)
{                      
  double lhs = (double)M[0]*(double)M[1];
  double rhs = (double)M[2]*(double)M[3];

  // Report result
  if (verbose)
    {
      cout << "Inequality " << ((lhs <= rhs) ? "holds " : "doesn't hold ");
      cout << "for "; G.report(); cout << endl;
      cout << "and choice of edges e = " << e << " and f = " << f << ". ";
      cout << endl << endl;

      cout << "M^{e, f}(G) * M_{e, f}(G) = " << M[0] << " * " << M[1] << " = " << lhs << " ";
      cout << ((lhs <= rhs) ? "<=" : ">"); 
      cout << " " << rhs << " = " << M[2] << " * " << M[3] << " = M^e_f(G) * M^f_e(G).";
      cout << endl << endl;
    }

  // Check correctness of M[k]
  if (check)
    {
      if (verbose) cout << "Checking result using conventional method: " << endl;
      G.compute_inequality(e, f, verbose, M);
    }

  // Report potential counterexample
  if (lhs > rhs)
    {
      cout << "Conjecture violated for "; G.report(); cout << endl;
      cout << "and choice of edges e = " << e << " and f = " << f << ". ";
      cout << endl << endl;

      return false;
    }

  // Record ratio rhs/lhs
  if (ratio != NULL) *ratio = rhs/lhs;

  return true;
}

//----------------------------------------------------------------------------------------------

// Compute inequality M^{e, f}(G) * M_{e, f}(G) <= M^e_f(G) * M^f_e(G)
// for edges e != f in E. Note: If cmpM != NULL, the obtained values are used to verify
// those provided in the array cmpM.

bool CGraph::compute_inequality(const CEdge& e, const CEdge& f, bool verbose, long *cmpM)
{
/*
  // Check input consistency: {e, f} subset E
  assert(has_edge(e));
  assert(has_edge(f));
*/
  // Compute M[k] for k = 0..3
  long M[4];

  // X_1 = {e, f}, Y_1 = {}
  CEdges X(e, f);
  CEdges Y;
  M[0] = count_forests(X, Y);

  // X_3 = {e}, Y_3 = {f}
  X.remove(f);
  Y.add(f);
  M[2] = count_forests(X, Y);

  // X_2 = {}, Y_2 = {e, f}
  X.remove(e);
  Y.add(e);
  M[1] = count_forests(X, Y);

  // X_4 = {f}, Y_4 = {e}
  X.add(f);
  Y.remove(f);
  M[3] = count_forests(X, Y);

  // Check inequality M[0] * M[1] <= M[2] * M[3]
  bool result = check_inequality(M, e, f, *this, NULL, verbose);

  if (cmpM != NULL)
    {
      // Verify values provided in the array cmpM
      for (int i = 0; i < 4; i++)
        {
          if (M[i] != cmpM[i]) 
            {             
              cout << endl <<endl;
              cout << "Provided value of M is incorrect!";
              cout << endl <<endl;
            }
        }
    }

  return result;
}

//----------------------------------------------------------------------------------------------

// Check conjecture M^{e, f}(G) * M_{e, f}(G) <= M^e_f(G) * M^f_e(G)
// for all pairs of distinct edges e, f in E. Note: This brute force approach is very slow!

bool CGraph::check_conjecture(bool verbose)
{
  // Output introductory information
  if (verbose)
    {
      cout << "Check inequality M^{e, f}(G) * M_{e, f}(G) <= M^e_f(G) * M^f_e(G) " << endl;
      cout << "for all pairs of distinct edges e, f in E, where G = (Y, E) is the ";
      report();
      cout << "." << endl << endl;
    }

  // Check conjecture for each pair of distinct edges e != f in G
  bool result = true;
  CEdge e_fail, f_fail;

  // Choose edge e
  for (int r1=0; r1<_n; r1++)
    {
      for (int c1=r1+1; c1<_n; c1++)
        {
          if (_A[r1][c1] == 1)
            {
              CEdge e(r1+1, c1+1);

              // Choose edge f
              for (int r2=r1; r2<_n; r2++)
                {
                  for (int c2=((r2==r1)?c1:r2)+1; c2<_n; c2++)
                    {
                      if (_A[r2][c2] == 1)
                        {                        
                          CEdge f(r2+1, c2+1);

                          // Check if inequality holds for conjecture instance (G, e, f)
                          if (!compute_inequality(e, f, verbose)) 
                            {
                              e_fail = e;
                              f_fail = f;
                              result = false;
                            }
                        }
                    }
                }
            }
        }
    }

  // Report result
  if (verbose)
    {
      if (result == false)
        {
          cout << "Conjecture disproved: " << endl;
          cout << "The inequality M^{e, f}(G) * M_{e, f}(G) <= M^e_f(G) * M^f_e(G) ";
          cout << "does not hold for ";
          report(); 
          cout << "and edges e = " << e_fail << ", f = " << f_fail << " in E.";
          cout << endl << endl;
        }
      else
        {
          cout << "Conjecture confirmed: " << endl;
          cout << "The inequality M^{e, f}(G) * M_{e, f}(G) <= M^e_f(G) * M^f_e(G) ";
          cout << "holds for all ";
          report();
          cout << "and all pairs of edges e != f in E.";
          cout << endl << endl;
        }
    }

  return result;
}

//----------------------------------------------------------------------------------------------

// Use the nauty-library to compute...

// 1. if I != NULL: C"(G, pi) for the partitioned graph (G, pi) with parition pi based on 
//                  I and delta: v in I has colour delta(v), all others are black (= 0).

// 2. if I == NULL: C"(G) for the (unpartitioned) graph G, also providing (if delta != NULL)
//                  the permutation delta = delta_G : G -> C"(G)

void CGraph::compute_isomorph(Graph_code *CG, CPermutation *delta, const vector<bool> *I) const
{
  // Assume that one setword suffices to store one row of the adjacency matrix 
  assert((_n + WORDSIZE - 1)/WORDSIZE == 1);
 
  // Allocate memory for nauty-computations
  DYNALLSTAT(graph, G, G_sz);
  DYNALLSTAT(graph, canonG, canonG_sz);
  DYNALLSTAT(nvector, lab, lab_sz);
  DYNALLSTAT(nvector, ptn, ptn_sz);
  DYNALLSTAT(nvector, orbits, orbits_sz);
  DYNALLSTAT(setword, workspace, workspace_sz);
  
  DYNALLOC1(graph, G, G_sz, _n, "CGraph::compute_isomorph");  
  DYNALLOC1(graph, canonG, canonG_sz, _n, "CGraph::compute_isomorph");
  DYNALLOC1(nvector, lab, lab_sz, _n, "CGraph::compute_isomorph");
  DYNALLOC1(nvector, ptn, ptn_sz, _n, "CGraph::compute_isomorph");
  DYNALLOC1(nvector, orbits, orbits_sz, _n, "CGraph::compute_isomorph");
  DYNALLOC1(setword, workspace, workspace_sz, 50*_n, "CGraph::compute_isomorph");

  DEFAULTOPTIONS(options);

  // Require canonically-labelled isomorph
  options.getcanon = TRUE;

  // Do not require automorphism group etc.
  options.writeautoms = FALSE;
  options.writemarkers = FALSE;

  statsblk stats;

  // Transcribe adjacency matrix A of G into nauty-format
  nset *row;
  for (int r=0; r<_n; r++)
    {
      row = GRAPHROW(G, r, 1);
      EMPTYSET(row, 1);
      for (int c=0; c<_n; c++)
        {    
          if (_A[r][c] == 1)
            {
              ADDELEMENT(row, c);
            }
        }
    }

  // Case 1. If I != NULL, compute colour classes / cells in nauty-format
  if(I != NULL)
    {
/*
      // Consistency checks
      assert(I->size() == _n);
      assert(delta != NULL);
      assert(delta->_image.size() == _n);
*/
      // Deactivate default monum_Chromatic colouring
      options.defaultptn = FALSE;

      // coloured_V[c-1] = v-1, where v is vertex in I with colour c = delta(v), or -1 if no such v exists
      // black_V[i-1] = v-1, where v is i-th black vertex, in increasing order
      // Note: Store v-1 instead of v because nauty's vertex labels range from 0 (not 1).
      vector<int> coloured_V(_n, -1);
      vector<int> black_V;
      int num_black = 0;

      // Extract colour classes
      for(int v=1; v<=_n; v++)
        {          
          if ((*I)[v-1])
            {
              coloured_V[delta->permute(v)-1] = v-1;
            }
          else
            {
              black_V.push_back(v-1);
              num_black++;
            }
        }
        
      // Transcribe black colour class
      int i;
      for (i=0; i<num_black; i++)
        {
          lab[i] = black_V[i];
          ptn[i] = (i == num_black-1) ? 0 : 1;
        }

      // Transcribe singleton colour classes for other colours
      int j = 0;
      for (i=num_black; i<_n; i++)
        {
          while (coloured_V[j] == -1) j++;
          lab[i] = coloured_V[j];
          ptn[i] = 0;
          j++;
        }
    }    

  // Execute nauty to determine the canonically-labelled isomorph of G
  nauty(G, lab, ptn, NILSET, orbits, &options, &stats, workspace, 50, 1, _n, canonG);

  // Transcribe the canonically-labelled isomorph to CG
/*
  assert(CG != NULL);
*/
  CG->resize(_n);
  int i;
  for (i=0; i<_n; i++)
    {
      (*CG)[i] = canonG[i];
    }

  // Case 2. If I == NULL and delta != NULL, transcribe canonical labelling permutation delta
  if ((delta != NULL) && (I == NULL))
    {
      delta->_image.resize(_n);

      for(int i=0; i<_n; i++)
        {
          delta->_image[lab[i]] = i+1;
        }
    }
/*
  // Deallocate memory
  DYNFREE(G, G_sz); 
  DYNFREE(canonG, canonG_sz);
  DYNFREE(lab, lab_sz)
  DYNFREE(ptn, ptn_sz);
  DYNFREE(orbits, orbits_sz);
  DYNFREE(workspace, workspace_sz);
*/
}

//----------------------------------------------------------------------------------------------

// Construct the set of components of the graph F. If F is acyclic, this will be the
// smallest forests containing F. Otherwise, abort as soon as a cycle is found.

bool CGraph::construct_forest(const vector<CEdge>& F, Component_set& C) const
{
  vector<bool> isolated(_n, true);
  CComponent *new_c;
  vector<CEdge>::const_iterator f;
   
  // Add all edges f in F to root forest
  for (f = F.begin(); f != F.end(); f++)
    {
      // Attempt to join edge f to existing component(s)
      vector<CComponent*> adj_C;
      int num_adj_C = 0;

      // Find components on which f is incident
      Component_set::iterator c;
      for (c = C.begin(); c != C.end() && num_adj_C < 2; c++)
        {
          int incidences = (*c)->_V.find_incidences(*f);

          if (incidences == 2)
            {
              // If both endvertices of f are incident with the 
              // present component, then F contains a cycle -> exit routine.
              return false;
            }
          else if (incidences == 1)
            {
              // If one endvertex of f is incident with the present
              // component, this component is shortlisted as adjacent to f.
              adj_C.push_back(*c);
              num_adj_C++;
            }
        }

      if (num_adj_C == 2)
        {
          // If f joins two components, then f merges them.
          new_c = new CComponent(adj_C[0], adj_C[1], *f);
          
          // Remove constitutent components
          assert(C.erase(adj_C[0]) == 1);
          assert(C.erase(adj_C[1]) == 1);
          delete adj_C[0];
          delete adj_C[1];
        }
      else if (num_adj_C == 1)
        {
          // If f is incident with exactly one component, then f enlarges it.
          new_c = new CComponent(*adj_C[0]);
          new_c->add_edge(*f);

          // Remove previous component
          assert(C.erase(adj_C[0]) == 1);
          delete adj_C[0];
        }
      else
        {
          // If f is not incident with any component, then f forms a new singleton component
          new_c = new CComponent(*f);  
        }

      // Record new component
      C.insert(new_c);

      // Mark tail and head of f as non-isolated
      isolated[f->_v-1] = false;
      isolated[f->_w-1] = false;
    }
  
  // Form components representing the remaining isolated vertices
  for (int v=0; v<_n; v++)
    {
      if (isolated[v])
        {
          C.insert(new CComponent(v+1));
        }
    }

  return true;
}

//----------------------------------------------------------------------------------------------

// Compute M^X_Y(G), the number of spanning forests in G containing X and disjoined from Y.
// More precisely, let F be an acyclic subgraph of G and let X, Y be subsets of E 
// such that X is contained in F and Y is disjoint from F. Then M^X_Y(G) gives the number 
// of distinct choices of F for given G, X and Y.

long CGraph::count_forests(const CEdges& X, const CEdges& Y, bool verbose)
{
  Component_set component_DB;
  Forest_set forest_DB[2];
  int X_size = X._E.size();  
  int Y_size = Y._E.size();
  int i, j;
/*
  // Consistency checks

  // X subset of E
  for (i=0; i<X_size; i++) assert(has_edge(X._E[i]));
  
  // Y subset of E
  for (i=0; i<Y_size; i++) assert(has_edge(Y._E[i]));
  
  // X and Y disjoint
  for (i=0; i<X_size; i++)
    {
      for (j=0; j<Y_size; j++)
        {
          assert(!(X._E[i] == Y._E[j]));
        }
    }
*/
  // Output introductory information
  if (verbose)
    {
      cout << "Counting all spanning forests in G " << endl;
      cout << "containing X = " << X << " and disjoint from Y = " << Y << "," << endl;
      cout << "where G is the "; report(); cout << ".";
      cout << endl << endl;
    }

  // Construct the set of components C of X. These form the first components to be recorded
  // in the component database. If X is found cyclic, return M^X_Y(G) = 0
  if (!construct_forest(X._E, component_DB)) return 0;

  // Record smallest spanning forest in G containing X, namely the forest with edge set X. 
  forest_DB[0].insert(CForest(component_DB));

  if (verbose)
    {
      cout << "Smallest ";
      forest_DB[0].begin()->report();
      cout << endl;
    }

  // Remove edges in Y from adjacency matrix (using M^X_Y(G) = M^X(D) with D = G - Y)
  for (i=0; i<Y_size; i++)
    {      
      assert(remove_edge(Y._E[i]));
    }

  // Iterate through all layers
  long num_F = 1;
  long num_F_s = 1;
  for (int s=1; s<_n; s++)
    {      
      // No forests have yet been found on the current layer
      forest_DB[s%2].clear();
      num_F_s = 0;

      // For each forest F on the lower level
      Forest_set::iterator F;
      for (F = forest_DB[(s-1)%2].begin(); F != forest_DB[(s-1)%2].end(); F++)
        {
          // For each pair of components of forest F on the lower layer
          int num_c = F->_C.size();
          for (i=0; i<num_c; i++)
            {
              for (j=i+1; j<num_c; j++)
                {
                  // For each edge of G joining the chosen pair of components
                  CVertices& V = F->_C[i]->_V;
                  CVertices& W = F->_C[j]->_V;
                  int num_v = V._V.size();
                  int num_w = W._V.size();

                  for (int r=0; r<num_v; r++)
                    {
                      int v = V._V[r];
                      for (int c=0; c<num_w; c++)
                        {
                          int w = W._V[c];
                          if (has_edge(v, w))
                            {   
                              // Contruct a new forest by adding the chosen edge, 
                              // joining the chosen pair of components
                              CEdge e(v, w);
                              CComponent *new_c = new CComponent(F->_C[i], F->_C[j], e);
                              
                              pair<Component_set::iterator, bool> result = component_DB.insert(new_c);
                              if (!result.second) delete new_c; 
                              
                              CForest new_F(*F, F->_C[i], F->_C[j], *result.first);
                              if ((forest_DB[s%2].insert(new_F)).second) num_F_s++;
                            }
                        }
                    }
                }
            }              
        }

      // Report spanning forests of size s
      if (verbose)
        {
          cout << "There are " << num_F_s << " spanning forests of size " << X_size + s << ": " << endl;

          long h = 0;
          for (F = forest_DB[s%2].begin(); F != forest_DB[s%2].end(); F++)
            {
              cout << ++h << ". ";
              F->report();
            }

          cout << endl;
        }

      num_F += num_F_s;         
    }

  // Clear database of components
  Component_set::iterator c;
  for (c = component_DB.begin(); c != component_DB.end(); c++)
    {
      delete *c;
    }

  // Restore edge set Y in adjacency matrix
  for (i=0; i<Y_size; i++)
    {
      assert(add_edge(Y._E[i]));
    }

  return num_F;
}

//----------------------------------------------------------------------------------------------

// Determine the smallest index j = j_max for which X_j is cyclic OR M^{X_j}(D) is available
// from the M-database (if neither holds for all j, return j = k). Cyclic indicates if X_j was 
// identified as cyclic, otherwise M_subtot provides M^{X_j}(D) if already available. I_sizes
// and indices give list of keys for values M^{X_i}(D), i < j, to be determined and stored.

int CGraph::compute_j_max(long& M_subtot, const CEdges& D_X, CEdges& X_j, CGraph& GX_j, CGraph& GD_j_X_j, 
                          CPermutation& delta_D, int s, vector<bool>& I, int& I_size, 
                          vector<int>& I_sizes, vector< M_index >& indices, M_maps& M_DB)
{
  // Initialize C with components of X
  CEdges& X = X_j;
  Component_set C;

  // If X is non-empty
  if (X.size() >= 1) 
    {
      // Start off with single edge e
      const CEdge& e = X._E[0];
      C.insert(new CComponent(e));

      // If X contains a second edge f
      if (X.size() == 2)
        {
          CComponent *c = *C.begin();
          CComponent *new_c;
          const CEdge& f = X._E[1];

          // Determine if f is incident with e
          if (c->_V.find_incidences(f))
            {
              // If so, f enlarges component {e} to form {e, f}
              new_c = new CComponent(*c);
              new_c->add_edge(f);

              // Remove previous component
              assert(C.erase(c) == 1);
              delete c;
            }
          else
            {
              // Otherwise, f forms a new singleton component {f}
              new_c = new CComponent(f);
            }

          // Record new component          
          C.insert(new_c);
        }
    }

  // Let D - X = {d_1, ..., d_k}. Keep adding edges d_j to X_j 
  // so long as it remains acylic
  bool cyclic = false;
  int j = 0;
  int k = D_X.size();
  CComponent *new_c;
  cyclic = false;

  do
    {
      // Increase j
      j++;

      // Consider edge g_j
      const CEdge& g_j = D_X._E[j-1];

      // Update sets X_j and D_j - X_j and corresponding graphs, updating interface I
      X_j.push(g_j);
      GX_j.add_edge(g_j);    
      GX_j.update_I_after_add(g_j, GD_j_X_j, I, I_size); 
      GD_j_X_j.remove_edge(g_j);      
      GD_j_X_j.update_I_after_remove(g_j, I, I_size); 

      // Check if M_subtot is already available
      if (obtain_M_subtot(M_subtot, GX_j, GD_j_X_j, delta_D, I, I_size, I_sizes, indices, M_DB)) break;      
      
      // Check if D - X is exhausted, i.e. recursion has reached M^D(D) to be computed
      if (j == k) break;

      // If storage-for-performance measures are actived
      if (s >= LOW_LAYER)
        {
          // Check if constraint X_j is deemed sufficiently large for M^{X_j}(D) to be efficiently
          // computed without further recursion
          if (X_j.size() >= SMALL_CONSTRAINT) break;
        }

      // Attempt to join edge g_j to existing component(s)
      vector<CComponent*> adj_C;
      int num_adj_C = 0;

      // Find components on which g_j is incident
      Component_set::iterator c;
      for (c = C.begin(); c != C.end() && num_adj_C < 2; c++)
        {
          int incidences = (*c)->_V.find_incidences(g_j);
          if (incidences == 2)                                                  
            {
              // If both endvertices of g_j are incident with the 
              // present component, then adding g_j forms a cycle in X_j.
              M_subtot = 0;
              cyclic = true;                                                  
              break;
            }
          else if (incidences == 1)
            {
              // If one endvertex of g_j is incident with the present
              // component, this component is shortlisted as adjacent to g_j.
              adj_C.push_back(*c);
              num_adj_C++;
            }
        }

      // Check if X_j has been identified as cyclic, so that M^{X_j}(D) = 0
      if (cyclic) break;

      if (num_adj_C == 2)
        {
          // If g_j joins two components, then g_j merges them.
          new_c = new CComponent(adj_C[0], adj_C[1], g_j);
       
          // Remove constitutent components
          C.erase(C.find(adj_C[0]));
          C.erase(C.find(adj_C[1]));
          delete adj_C[0];
          delete adj_C[1];
        }
      else if (adj_C.size() == 1)
        {
          // If g_j is incident with exactly one component, then g_j enlarges it.
          new_c = new CComponent(*adj_C[0]);
          new_c->add_edge(g_j);

          // Remove previous component
          assert(C.erase(adj_C[0]) == 1);
          delete adj_C[0];
        }
      else
        {
          // If g_j is not incident with any component, then g_j forms a new singleton component.
          new_c = new CComponent(g_j);
        }    

      // Record new component
      C.insert(new_c);
      
    } while (true);

  // Discard components
  Component_set::iterator c;
  for (c = C.begin(); c != C.end(); c++) delete *c;

  return j;
}

//----------------------------------------------------------------------------------------------

// Store M^{X_j}(D) = M_subtot in the M-database, under the given triple 
// < |I|, index >, where index = < C"(X_j, pi), C"(D - X_j, pi) >

void CGraph::store_M_subtot(long M_subtot, int I_size, M_index& index, M_maps& M_DB)
{
  M_DB[I_size].insert(M_entry(index, CM_value(M_subtot, 2)));
}

//----------------------------------------------------------------------------------------------

// Obtain M^{X_j}(D) = M_subtot from the M-database ONLY IF this value is already available;
// otherwise record |I| and < C"(X_j, pi), C"(D - X_j, pi) >. Return if retrieval succeeded.

bool CGraph::obtain_M_subtot(long& M_subtot, const CGraph& GX_j, const CGraph& GD_X_j,
                             CPermutation& delta_D, const vector<bool>& I, int I_size, 
                             vector<int>& I_sizes, vector< M_index >& indices, M_maps& M_DB)
{
  // Static variables for look-up in M-database
  static Graph_code CX_j, CD_X_j;
  static CEdges emptyset;

  GX_j.compute_isomorph(&CX_j, &delta_D, &I);
  GD_X_j.compute_isomorph(&CD_X_j, &delta_D, &I);
  M_index index(make_pair(CX_j, CD_X_j));
 
  // Try to retrieve value of M_subtot from M-database
  M_map::iterator result = M_DB[I_size].find(index);
  if (result == M_DB[I_size].end()) 
    {
      // Record |I| and < C"(X_j, pi), C"(D - X_j, pi) > to store M_subtot later
      I_sizes.push_back(I_size);
      indices.push_back(index);    
      return false;
    }

  // Entry available: retrieve M_subtot
  M_subtot = result->second._M;
  return true;
}

//----------------------------------------------------------------------------------------------

// Try to retrieve M_minor from the M-database; if not available, 
// compute and store M_minor for later use. Return the obtained value.

long CGraph::obtain_M_minor(const CEdges& X_j, const CGraph& GX_j, const CGraph& GD_j_X_j,                         
                            CGraph& D_j, CPermutation& delta_D_j, const vector<bool>& I, int I_size, 
                            int s, M_maps& M_DB)
{
  // Static variables for look-up in M-database
  static Graph_code CX_j, CD_j_X_j;
  static CEdges emptyset;

  long M_minor;

  // Initialise index < |I|, C"(X_j, pi), C"(D_j - X_j, pi) >
  GX_j.compute_isomorph(&CX_j, &delta_D_j, &I);
  GD_j_X_j.compute_isomorph(&CD_j_X_j, &delta_D_j, &I);
  M_index index(make_pair(CX_j, CD_j_X_j));

  // Try to retrieve value of M_minor from M-database
  M_map::iterator result = M_DB[I_size].find(index);
  if (result == M_DB[I_size].end())
    {
      // No entry available for the given index: compute and store M_minor
      M_minor = D_j.count_forests(X_j, emptyset, false);
      M_DB[I_size].insert(M_entry(index, CM_value(M_minor, 3)));
    }
  else
    {
      // Entry available: retrieve M_minor
      M_minor = result->second._M;
    }

  return M_minor;
}

//----------------------------------------------------------------------------------------------

// Compute M^X(D) (for a detailed description see the documentation)

long CGraph::compute_M(CEdges &X, CGraph& GX, const CEdges& D_X, CGraph& GD_X, CGraph& GD, 
                       Graph_code& CD, CPermutation& delta_D, M_maps *M_DB, int n, int s)
{  
  // Initial value of X_j is X
  CEdges& X_j = X;
  CGraph& GX_j = GX;

  // Initial value of D_j - X_j is D - X
  CGraph& GD_j_X_j = GD_X;

  // Initialize interface I of X_j and D_j - X_j
  vector<bool> I;
  int I_size;
  GX_j.mark_I(GD_j_X_j, I, I_size);

  // Determine the largest index j = j_max up to which recursion will be performed and 
  // obtain M^{X_j}(D) if possible
  long M_subtot = -1;
  vector< M_index > indices;
  vector<int> I_sizes;
  int j_max = compute_j_max(M_subtot, D_X, X_j, GX_j, GD_j_X_j, delta_D, s, I, I_size, I_sizes, indices, M_DB[s%4]);

  // Initialize variables
  Graph_code CD_j;
  CPermutation delta_D_j(n);
  long M_tot = 0;
  int j = j_max;
  
  // Initialize M_tot by M^{X_{j_max}}(D), to be computed if not previously available
  if (M_subtot == -1) M_tot += obtain_M_minor(X_j, GX_j, GD_j_X_j, GD, delta_D, I, I_size, s, M_DB[s%4]);
  else                M_tot += M_subtot;

  // Obtain summands M_minor for j in {j_max, ..., 1} in decreasing order
  for (j = j_max; j >= 1; j--)
    {
      // Remove g_j from D to obtain D_j
      const CEdge& g_j = D_X._E[j-1];
      GD.remove_edge(g_j);
      CGraph& D_j = GD;

      // Compute C"(D_j) and delta_D_j
      D_j.compute_isomorph(&CD_j, &delta_D_j);

      // Update X_j by removing g_j (note: D_j - X_j remains unchanged)
      X_j.pop();
      GX_j.remove_edge(g_j);
      GX_j.update_I_after_remove(g_j, I, I_size);
        
      // Retrieve or compute M_minor and update M_tot
      M_tot += obtain_M_minor(X_j, GX_j, GD_j_X_j, D_j, delta_D_j, I, I_size, s-1, M_DB[(s-1)%4]);
      
      if (j > 1)
        {
          // Add g_j to D_j - X_j to obtain D - X_j
          GD_j_X_j.add_edge(g_j);
          GD_j_X_j.update_I_after_add(g_j, GX_j, I, I_size);
          CGraph& GD_X_j = GD_j_X_j;

          // Store M^{X_j}(D) = M_tot
          store_M_subtot(M_tot, I_sizes[j-2], indices[j-2], M_DB[s%4]);
        }

      // Add g_j to D_j to recover D
      GD.add_edge(g_j);
    }

  // The final sum is the required value M^X(D)
  return M_tot;
}

//----------------------------------------------------------------------------------------------

// Try to retrieve M^X(D) from the M-database; if not available, compute it 
// explicitly and buffer it for later use. If applicable, report whether the value 
// originates from checking an earlier equivalent conjecture instance.

bool CGraph::obtain_M(long& M, CGraph& G, const Graph_code& CG, const CPermutation& delta_G, 
                      CEdges& E, CEdges &X, CEdges &Y, M_maps *M_DB, int n, int s, Method method)
{
  // Static variables for look-up in M-database
  static Graph_code CX, CD_X;
  static CEdges emptyset;

  // If X = {e, f}, Y = {} are standard constraints and M^X_Y(G) is available from the database,
  // this flag indicates whether the retrieved value originates from checking an earlier 
  // conjecture instance equivalent to (G, e, f). If not applicable, set to false.
  bool earlier_equivalent_instance_checked = false;

  // Graph D = G - Y
  CGraph& GD = G;
  GD.remove_edges(Y);

  // Graph X
  CGraph GX(n, X);

  // Edge set D - X
  CEdges D_X(E);
  D_X.remove(Y);
  D_X.remove(X);

  // Graph D - X
  CGraph GD_X(GD);
  GD_X.remove_edges(X);

  // Obtain interface I(X, D - X) and its size |I|
  vector<bool> I;
  int I_size;
  GX.mark_I(GD_X, I, I_size);

  // Obtain C"(D) and delta_D
  Graph_code CD;
  CPermutation delta_D(n);
  if (Y.size() == 0) 
    {
      // If Y = {}, we have C"(D) = C"(G), delta_D = delta_G
      CD = CG;
      delta_D = delta_G;
    }
  else
    {
      // Otherwise compute  C"(D) and delta_D
      GD.compute_isomorph(&CD, &delta_D);
    }

  // Initialise index < |I|, C"(X, pi), C"(D - X, pi) >
  GX.compute_isomorph(&CX, &delta_D, &I);
  GD_X.compute_isomorph(&CD_X, &delta_D, &I);
  M_index index(make_pair(CX, CD_X));

  // Try to retrieve value of M^X(D) from M-database
  M_maps& M_DB_layer = M_DB[(s - Y.size())%4];
  M_map::iterator result = M_DB_layer[I_size].find(index);
  if (result == M_DB_layer[I_size].end())
    {
      // No entry available for the given index: compute and store M^X(D)
      switch(method)
        {
        case naive:

          // Count spanning forests directly
          M = GD.count_forests(X, emptyset);
          break;
   
        case advanced:

          // For a detailed description see the documentation.
          M = compute_M(X, GX, D_X, GD_X, GD, CD, delta_D, M_DB, n, s - Y.size());
          break;
        }

      // Store the obtained value
      M_DB_layer[I_size].insert(M_entry(index, CM_value(M, 1)));
    }
  else
    {
      // Entry available: retrieve M^X(D)
      CM_value& M_value = result->second;
      M = M_value._M;
      earlier_equivalent_instance_checked = (M_value._origin == 1);
    }

  // Restore G
  G.add_edges(Y);

  return earlier_equivalent_instance_checked;
}

//----------------------------------------------------------------------------------------------

// Compute or retrieve the values M_k for k = 1, ..., 4 and verify the conjectured
// inequality M_1*M_2 <= M_3*M_4, unless an instance (G', e_i, f_i) ~ (G, e_i, f_i) 
// has been checked before; return 'true' if this occurs.

bool CGraph::check_instance(CGraph& G, const Graph_code& CG, const CPermutation& delta_G, 
                            CEdges& E, const CEdge& e, const CEdge& f, CEdges X[4], CEdges Y[4], 
                            M_maps *M_DB, int n, int s, CRatiostats& stats, 
                            bool& conjecture, Method method, bool verbose, bool check)
{
  // Values of M_k to be computed or retrieved
  long M[4];

  // Obtain M_1 = M^{e, f}(G)
  if (obtain_M(M[0], G, CG, delta_G, E, X[0], Y[0], M_DB, n, s, method)) 
    {
      // If the value of M^{e, f}(G) is successfully retrieved, we conclude 
      // that some equivalent instance (G', e, f) has been checked before -> skip (G, e, f)
      return true;
    }

  // Otherwise, (G, e, f) is idenfied as new instance ->
  // retrieve or compute M_2 = M^{e, f}(G), M_3 = M^e_f(G) and M_4 = M^f_e.
  for (int k=1; k<4; k++) 
    {
      obtain_M(M[k], G, CG, delta_G, E, X[k], Y[k], M_DB, n, s, method);
    }

  // Now check the conjectured inequality M_1*M_2 <= M_3*M_4; 
  // recording the ratio rhs/lhs and verifying the correctness of M_k, if desired.
  double ratio;
  check_inequality(M, e, f, G, &ratio, verbose, check);
  stats.update(s, ratio, G);

  // Update the count of conjecture instances analyzed
  if (!verbose && (stats._num_instances[s]%1000 == 1)) cout << ".";

  return false;
}

//----------------------------------------------------------------------------------------------


// Generate supergraphs G of subG, by adding one absent edge each, and verify the conjectured
// inequality for all (non-equivalent) conjecture instances (G, e_i, f_i), for i=1,2

void CGraph::analyse_supergraphs(const CG_entry& subG_entry, CEdge e[2], CEdge f[2], 
                                 CEdges X[2][4], CEdges Y[2][4], Graph_list *graph_DB, M_maps *M_DB,
                                 int n, int s, int i_max, CRatiostats **stats, bool& conjecture,
                                 Method method, bool verbose, bool check)
{
  // Reinstantiate subG form its graph code
  CGraph subG(subG_entry._code);
  int skip_G = subG_entry._skip;

  // Obtain the corresponding edge set E in lex order
  CEdges E;
  subG.get_edges(E);

  // Determine the first isolated vertex v_star
  int v_star = E.first_isolated_vertex();

  // We will examine conjecture instances (subG + g, e_i, f_i) for those i, 
  // for which E_i = { e_i, f_i } is already contained in subG.
  bool G_contains_E_i[2];
  for (int i=0; i<i_max; i++)
    {                
      assert(subG.has_edge(e[i]));
      G_contains_E_i[i] = subG.has_edge(f[i]);
    }

  // Consider all choices of edges not present in subG in lex order.
  // It is enough to add edges <v, w> with v, w <= first isolated vertex v_star.
  for (int v=1; v<=v_star; v++)
    {
      int w_max = min((v == v_star) ? (v_star + 1) : v_star, n);
      for (int w=v+1; w<=w_max; w++)
        {
          CEdge g(v, w);
      
          if (!subG.has_edge(g))
            { 
              // If g = <v, w> is not present in subG, we add it to form G = subG + g.
              // Note: We intentionally put g last in E.
              E.push(g);
              subG.add_edge(g);
              CGraph& G = subG;

              // Compute C"(G) and delta_G : G -> C"(G)
              Graph_code CG;
              CPermutation delta_G(n);
              G.compute_isomorph(&CG, &delta_G);

              // For each E_i = { e_i, f_i }
              bool skipped[2] = { false, false };
              bool new_instance = false;
              for (int i=0; i<i_max; i++)
                {                
                  // Check conjecture instance (G, e_i_, f_i) only if G contains E_i
                  // and, for all ancestors H of G, no (H, e_i, f_i) was ever found equivalent 
                  // to any (H', e_i, f_i), which would be indicated by skip_G != i.
                  if (G_contains_E_i[i] && (skip_G != i))
                    {
                      // Check conjecture instance (G, e_i, f_i). When the method returns true, we have 
                      // (G, e_i, f_i) ~ (G', e_i, f_i) for some G', thus (H, e_i, f_i) ~ (H', e_i, f_i) 
                      // for descendants H, H' of G, G' respectively -> may skip such H.
                      skipped[i] = check_instance(G, CG, delta_G, E, e[i], f[i], X[i], Y[i], M_DB,
                                                  n, s, *stats[i], conjecture, method, verbose, check);

                      // Record if any of the two instances (G, e_i, f_i) with i = 1, 2 is identified 
                      // as new, i.e. if it is worth recording G for the later analysis its descendants H.
                      if (!skipped[i]) new_instance = true;
                    }
                }

              // If (G, e_i, f_i) forms a new conjecture instance for either (or both) i = 1 or 2, 
              if (new_instance)
                {
                  // Record G for the later analysis of its descendants H.
                  Graph_code G_code;
                  G.get_graph_code(G_code);

                  // Record for which (if any) i we may skip checking (H, e_i, f_i)
                  int skip_H;
                  if (skip_G != -1)     skipped[skip_G] = true;
                  if (skipped[0])       skip_H = 0;
                  else if (skipped[1])  skip_H = 1;
                  else                  skip_H = -1;

                  graph_DB[s%2].push_back(CG_entry(G_code, skip_H));
                }

              // Remove g from G to recover subG
              E.pop();
              subG.remove_edge(g);
            }
        }
    }
}

//----------------------------------------------------------------------------------------------

// Check inequality M^{e, f}(G) * M_{e, f}(G) <= M^e_f(G) * M^f_e(G)
// for all distinct conjecture instances (G, e, f) with G contained in K_n and e != f in E(G).

bool CGraph::check_conjecture_up_to_K_n(int n, Method method, bool verbose, bool check)
{
  bool conjecture = true;

  // Output introductory information
  cout << "Checking conjecture ";
  cout << "M^{e, f}(G) * M_{e, f}(G) <= M^e_f(G) * M^f_e(G)" << endl;
  cout << "for all distinct conjecture instances (G, e, f), ";
  cout << "with G contained in M_" << n << " and edges e != f in E(G)";
  cout << endl << endl;

  // For graphs G = (V, E) on <= 2 vertices, the conjecture is meaningless. 
  if (n <= 2) 
    { 
      cout << "Conjecture meaningless for n <= 2." << endl; 
      return true; 
    }

  // Generate edge sets E_1 = {e_1, f_1} = {<1, 2>, <1, 3>} and, for n >= 4, 
  // E_2 = {e_2, f_2} = {<1, 2>, <3, 4>}, the only two non-isomorphic choices of e != f.
  vector<CEdges> E_i; 
  E_i.push_back(CEdges(2*2, 1, 2, 1, 3));
  if (n >= 4) E_i.push_back(CEdges(2*2, 1, 2, 3, 4));
  int i, i_max = E_i.size();

  // We enumerate the hierarchy of supergraphs of E_i up to K_n; there are N = (n choose 2) 
  // layers to examine (not counting the 0-th layer, which consists of the empty graph only).
  int s;
  int N = n*(n-1)/2;

  cout << "There are " << E_i.size() << " isomorphically distinct choices of {e, f} ";
  cout << "and (" << n << " choose 2) = " << N << " graph layers to examine. ";
  cout << endl << endl;
       
  // Graph database storing two layers at a time
  Graph_list graph_DB[2];

  // Database storing M^X(D), for four layers of graphs D at a time,
  // indexed by the triple (|I|, C"(X, pi), C"(D - X, pi))
  M_maps M_DB[4];
  for (i=0; i < 4; i++) M_DB[i].resize(n+1);

  // Initialisations 
  CEdge e[2], f[2];
  CEdges X[2][4], Y[2][4], emptyset;
  CRatiostats* stats[2];

  for (i=0; i<i_max; i++)
    {
      // Initialise...

      // -  Graph database with { E_1, E_2 } on layer 2.
      //    Note: As may be verified directly, the conjecture holds for these.
      CGraph GE_i(n, E_i[i]);    
      Graph_code GE_i_code;
      GE_i.get_graph_code(GE_i_code);
      graph_DB[0].push_back(CG_entry(GE_i_code, -1));

      // -  Edges e_i, f_i
      e[i] = E_i[i]._E[0];     
      f[i] = E_i[i]._E[1];
      
      // -  Edge sets X_{i, k} = {e_i, f_i} - Y_{i, k} =
      //    {e_i, f_i}, {}, {e_i} and {f_i} for k = 1, 2, 3, 4, respectively
      X[i][0] = E_i[i];           Y[i][0] = emptyset;
      X[i][1] = emptyset;         Y[i][1] = E_i[i];
      X[i][2] = CEdges(e[i]);     Y[i][2] = CEdges(f[i]);
      X[i][3] = CEdges(f[i]);     Y[i][3] = CEdges(e[i]);

      // -  Ratio statistics
      double res = (i == 0) ? 0.1 : 0.02;
      stats[i] = new CRatiostats(n, N+1, res, 1.0, i);
    }

  // For all layers s from 3 to N
  for (s=3; s<=N; s++)
    {      
      // Report current layer
      cout << "Layer " << s << ": " << endl;

      // Clear current layer of both graph and M-database
      graph_DB[s%2].clear();
      for (i=0; i <= n; i++) M_DB[s%4][i].clear();

      // For all graphs G stored on the layer below
      Graph_list::iterator G;
      for (G = graph_DB[(s-1)%2].begin(); G != graph_DB[(s-1)%2].end(); G++)
        {        
          // Analyse new conjecture instances (H, e_i, f_i) for subgraphs H of G
          analyse_supergraphs(*G, e, f, X, Y, graph_DB, M_DB, n, s, i_max, 
                              stats, conjecture, method, verbose, check);
        }

      // Report results
      cout << endl;
      for (i=0; i<E_i.size(); i++)
        {
          cout << "Analysis of " << stats[i]->_num_instances[s] << " distinct conjecture instances ";
          cout << "of size " << s << " with {e, f} = " << E_i[i] << " completed. ";
          cout << endl << endl;
          stats[i]->report(s);
          cout << endl << endl;
        }
    }

  // Plot and delete statistics
  for (i=0; i<i_max; i++)
    {
      stats[i]->plot(i);
      cout << endl << endl;
      delete stats[i];
    }

  return conjecture;
}

//---------------------------------------------------------------------------------------------------------------------------
