//*********************************************************************************************
// graph.h
//*********************************************************************************************

#ifndef __GRAPH_H__
#define __GRAPH_H__

#pragma message("Compiling graph.h\n")

// Required header files
#include "ratiostats.h"
#include "forest.h"
#include "graphutils.h"

// Externally defined classes
class CPermutation;

// Classes defined in this file
class CGraph;

//----------------------------------------------------------------------------------------------

// User-defined types
typedef vector< vector<int> > Matrix;
typedef enum { naive, improved, advanced } Method;

//----------------------------------------------------------------------------------------------

// Graph G = (V, E) with V = [n] and defined via adjacency matrix A
class CGraph
{
 private:
  friend ostream& operator<<(ostream& os, const CGraph& G);

  // Order n = |V|
  int _n;

  // Adjacency matrix, A_ij = 1 :<=> {i, j} in E
  Matrix _A;

  
  CGraph(const Graph_code& G);

  
  Graph_code& get_graph_code(Graph_code& G);

  CEdges& get_edges(CEdges& E);


  void find_V(vector<bool>& VG) const;

  void mark_I(CGraph& H, vector<bool>& I, int& I_size) const;

  bool update_I_after_add(int v, CGraph& H, vector<bool>& I) const; 

  bool update_I_after_remove(int v, vector<bool>& I) const;

  bool update_I_after_add(const CEdge& e, CGraph& H, vector<bool>& I, int& I_size) const;
  
  bool update_I_after_remove(const CEdge& e, vector<bool>& I, int& I_size) const;

  static bool check_inequality(long *M, const CEdge& e, const CEdge& f, CGraph& G, 
                               double *ratio = NULL, bool verbose = false, bool check = false);

  void compute_isomorph(Graph_code *CG, CPermutation *delta = NULL, 
                        const vector<bool> *I = NULL) const;

  bool construct_forest(const vector<CEdge>& F, Component_set& C) const;

  static int compute_j_max(long& M_subtotal, const CEdges& E, CEdges& X_j, CGraph& GX_j, CGraph& GD_j_X_j, 
                           CPermutation& delta_D, int s, vector<bool>& I, int& I_size, 
                           vector<int>& I_sizes, vector< M_index >& indices, M_maps& M_DB);

  static void store_M_subtot(long M_subtot, int I_size, M_index& index, M_maps& M_DB);

  static bool obtain_M_subtot(long& M_subtot, const CGraph& GX_j, const CGraph& GD_X_j,
                              CPermutation& delta_D, const vector<bool>& I, int I_size, 
                              vector<int>& I_sizes, vector< M_index >& indices, M_maps& M_DB);

  static long obtain_M_minor(const CEdges& X_j, const CGraph& GX_j, const CGraph& GD_j_X_j, CGraph& D_j, 
                             CPermutation& delta_D_j, const vector<bool>& I, int I_size, int s, M_maps& M_DB);

  static long compute_M(CEdges &X, CGraph& GX, const CEdges& D_X, CGraph& GD_X, 
                        CGraph& GD, Graph_code& CD, CPermutation& delta_D, M_maps *M_DB, int n, int s);

  static bool obtain_M(long& M, CGraph& G, const Graph_code& CG, const CPermutation& delta_G, 
                       CEdges& E, CEdges &X, CEdges &Y, M_maps *M_DB, int n, int s, Method method);

  static bool check_instance(CGraph& G, const Graph_code& CG, const CPermutation& delta_G, 
                             CEdges& E, const CEdge& e, const CEdge& f, CEdges X[4], CEdges Y[4], 
                             M_maps *M_DB, int n, int s, CRatiostats& stats, 
                             bool& conjecture, Method method, bool verbose, bool check);

  static void analyse_supergraphs(const CG_entry& subG_entry, CEdge e[2], CEdge f[2], 
                                  CEdges X[2][4], CEdges Y[2][4], Graph_list *graph_DB, M_maps *M_DB,
                                  int n, int s, int i_max, CRatiostats **stats, bool& conjecture,
                                  Method method, bool verbose, bool check);
 
 public:

  CGraph() {}

  CGraph(int n, const CEdges& E);

  CGraph(int n, int s);

  CGraph(int n, double p);


  bool has_edge(int i, int j) const;

  bool has_edge(const CEdge& e) const;

  bool add_edge(const CEdge& f);

  int add_edges(const CEdges& F);

  bool remove_edge(const CEdge& d);

  int remove_edges(const CEdges& D);


  void sample_random_subgraph(int m, CEdges& D);
  
  void report(void) const;


  long count_forests(const CEdges& X, const CEdges& Y, bool verbose = false);
  
  bool compute_inequality(const CEdge& e, const CEdge& f, bool verbose = false, long *cmpM = NULL);
 
  bool check_conjecture(bool verbose = false);
 
  static bool check_conjecture_up_to_K_n(int n, Method method = advanced, bool verbose = false, bool check = false);
};

ostream& operator<<(ostream& os, const CGraph& G);

//----------------------------------------------------------------------------------------------

#endif
