//*********************************************************************************************
// graphutils.cpp
//*********************************************************************************************

#include "StdAfx.h"

#include "graph.h"

//----------------------------------------------------------------------------------------------

// Construct G from its graph code
CGraph::CGraph(const Graph_code& G_code)
{
  CEdges E;
  _n = G_code.size();
  _A.resize(_n);

  DYNALLSTAT(graph, G, G_sz);
  DYNALLOC1(graph, G, G_sz, _n, "CGraph::CGraph");  

  for (int i=0; i<_n; i++)
    {
      G[i] = G_code[i];
    }

  nset *row;
  for (int r=0; r<_n; r++)
    {
      row = GRAPHROW(G, r, 1);
      _A[r].resize(_n);
      for (int c=0; c<_n; c++)
        {    
          if (ISELEMENT(row, c)) _A[r][c] = 1;
          else _A[r][c] = 0;
        }
    }

//DYNFREE(G, G_sz); 
}

//----------------------------------------------------------------------------------------------

// Obtain the graph code representing G
Graph_code& CGraph::get_graph_code(Graph_code& G_code)
{
  // Assume that one setword suffices to store one row of the adjacency matrix 
  assert((_n + WORDSIZE - 1)/WORDSIZE == 1);
 
  // Sllocate memory for nauty-computations
  DYNALLSTAT(graph, G, G_sz); 
  DYNALLOC1(graph, G, G_sz, _n, "CGraph::get_graph_code");  

  // Transcribe adjacency matrix A of G into nauty-format
  nset *row;
  for (int r=0; r<_n; r++)
    {
      row = GRAPHROW(G, r, 1);
      EMPTYSET(row, 1);
      for (int c=0; c<_n; c++)
        {    
          if (_A[r][c] == 1)
            {
              ADDELEMENT(row, c);
            }
        }
    }
	
  G_code.resize(_n);
  for (int i=0; i<_n; i++)
    {
      G_code[i] = G[i];
    }

//DYNFREE(G, G_sz); 

  return G_code;
}

//----------------------------------------------------------------------------------------------

// Obtain the edge set E of G
CEdges& CGraph::get_edges(CEdges& E)
{
  for (int r=1; r<=_n; r++)
    {
      for (int c=r+1; c<=_n; c++)
        {    
          if (has_edge(r, c))
            {
              E.push(CEdge(r, c));
            }
        }
    }

  return E;
}

//----------------------------------------------------------------------------------------------

// Find the set of non-isolated vertices V(G) of G
void CGraph::find_V(vector<bool>& VG) const
{
/*
  // Assert VG is initially empty
  assert(VG.size() == 0);
*/
  VG.resize(_n, false);

  for (int r=1; r<=_n; r++)
    {
      for (int c=1; c<=_n; c++)
        {
          if (has_edge(r, c))
            {
              VG[r-1] = true;
              break;
            }
        }
    }
}

//----------------------------------------------------------------------------------------------

// Mark the interface I(G, H) of this graph G and graph H
void CGraph::mark_I(CGraph& H, vector<bool>& I, int& I_size) const
{
/*
  // Assert I is initially empty
  assert(I.size() == 0);
*/
  // Find non-isolated vertices of G and H
  vector<bool>& VG = I;
  find_V(VG);
  vector<bool> VH;
  H.find_V(VH);
  
  // Mark the interface I(G, H)
  I_size = 0;
  for (int i=0; i<_n; i++)
    {
      I[i] = (VG[i] && VH[i]);
      if (I[i]) I_size++;
    }
}

//----------------------------------------------------------------------------------------------

// Update the interface I(G, H) after adding an edge incident with vertex v to G
bool CGraph::update_I_after_add(int v, CGraph& H, vector<bool>& I) const
{
  bool change = false;

  // Check if v is also connected in H
  for (int c=1; c<=_n; c++)
    {
      if (H.has_edge(v, c)) 
        {
          change = true;
          break;
        }
    }

  if (change)
    {
      change = !I[v-1];
      I[v-1] = true;
    }

  return change;
}

//----------------------------------------------------------------------------------------------

// Update the interface I(G, H) after removing an edge incident with vertex v from G
bool CGraph::update_I_after_remove(int v, vector<bool>& I) const
{
  bool change = true;

  // Check if v is still connected in G after removing edge
  for (int c=1; c<=_n; c++)
    {
      if (has_edge(v, c)) 
        {
          change = false;
          break;
        }
    }

  if (change)
    {
      change = I[v-1];
      I[v-1] = false;
    }

  return change;
}

//----------------------------------------------------------------------------------------------

// Update the interface I(G, H) after adding edge e to G
bool CGraph::update_I_after_add(const CEdge& e, CGraph& H, vector<bool>& I, int& I_size) const
{
  bool change1 = false, change2 = false;
  if (!I[e._v-1]) change1 = update_I_after_add(e._v, H, I);
  if (!I[e._w-1]) change2 = update_I_after_add(e._w, H, I);

  if (change1) I_size++;
  if (change2) I_size++;

  return (change1 || change2);
}

//----------------------------------------------------------------------------------------------

// Update the interface I(G, H) after removing edge e from G
bool CGraph::update_I_after_remove(const CEdge& e, vector<bool>& I, int& I_size) const
{
  bool change1 = false, change2 = false;
  if (I[e._v-1]) change1 = update_I_after_remove(e._v, I);
  if (I[e._w-1]) change2 = update_I_after_remove(e._w, I);

  if (change1) I_size--;
  if (change2) I_size--;

  return (change1 || change2);
}

//----------------------------------------------------------------------------------------------

// Construct a graph on n vertices with specified edge set 
CGraph::CGraph(int n, const CEdges& E) 
{ 
  _n = n;
  _A.resize(_n, vector<int>(_n, 0));

  int s = E.size();
  for (int i=0; i<s; i++)
    {
      assert(add_edge(E._E[i]));
    }
}


//----------------------------------------------------------------------------------------------

// Sample a random graph of size s on n vertices 
// with edges appearing independently and uniformly at random
CGraph::CGraph(int n, int s) : _A(n, vector<int>(n, 1))
{
  _n = n;
  
  // A initially complete, without loops
  for (int i=0; i<_n; i++) _A[i][i] = 0;

  // Remove s edges uniformly at random
  CEdges E;
  sample_random_subgraph(s, E);
  remove_edges(E);

  // Invert off-diagonal entries
  for (int r=0; r<_n; r++)
    {
      for (int c=0; c<_n; c++)
        {    
          if (r != c) _A[r][c] = _A[r][c] ? 0 : 1;
        }
    }
}

//----------------------------------------------------------------------------------------------

// Sample a random graph on n vertices 
// with edges appearing independently and uniformly at random with probability p
CGraph::CGraph(int n, double p) : _A(n, vector<int>(n, 0))
{
  _n = n;
  
  // G initially empty
  assert(p >= 0 && p <= 1);
  for (int i=0; i<_n; i++)
    {
      for (int j=0; j<_n; j++)
        {         
          if (j < i) 
            {
              // A is symmetric: copy upper diagonal entry
              _A[i][j] = _A[j][i];
            }
          else if (j > i)
            {     
              // Edge is present with probability p
              if ((double)rand()/(double)RAND_MAX < p)
                { 
                  _A[i][j] = 1;
                }
            }
        }
    }
}

//----------------------------------------------------------------------------------------------

// Sample uniformly at random a subset D of size min(m, |E|) of E
void CGraph::sample_random_subgraph(int m, CEdges& D)
{
  // Obtain list of edges E from adjacency matrix A
  vector<CEdge*> E;
  for (int r=1; r<=_n; r++)
    {
      for (int c=r+1; c<=_n; c++)
        {    
          if (has_edge(r, c)) E.push_back(new CEdge(r, c));
        }
    }

  // Size of random subset D is bounded by |E|
  int s = E.size();
  m = min(m, s);

  // Sample elements of D
  int i, j, k;
  D._E.resize(m);
  for (i=0; i<m; i++)
    {
      // Randomly choose one of the edge remaining in E - D
      j = rand()%s;
      
      // Add this edge to D
      D._E[i] = *(E[j]);
      
      // Remove this edge from E - D
      delete E[j];
      for (k=j+1; k<s; k++) E[k-1] = E[k];
      s--;
    }

  // Discard remaining edges in E - D
  for (k=0; k<s; k++) delete E[k];
}

//----------------------------------------------------------------------------------------------

// Check if edge e = {i, j} is present in G
bool CGraph::has_edge(int i, int j) const
{
  if (_A[i-1][j-1] == 1)  
    {
/*
      assert(_A[j-1][i-1] == 1);
*/
      return true;
    }
/*
  assert(_A[i-1][j-1] == 0);
  assert(_A[j-1][i-1] == 0);
*/
  return false;
}

//----------------------------------------------------------------------------------------------

// Check if edge e is present in G
bool CGraph::has_edge(const CEdge& e) const
{
  if (_A[e._v-1][e._w-1] == 1)  
    {
/*
      assert(_A[e._w-1][e._v-1] == 1);
*/
      return true;
    }
/*
  assert(_A[e._v-1][e._w-1] == 0);
  assert(_A[e._w-1][e._v-1] == 0);
*/
  return false;
}

//----------------------------------------------------------------------------------------------

// Add edge f to G; return if f was not previously present in G.
bool CGraph::add_edge(const CEdge& f)
{
  if (_A[f._v-1][f._w-1] == 0)
    {
/*
      assert(_A[f._w-1][f._v-1] == 0);
*/
      _A[f._v-1][f._w-1] = 1;
      _A[f._w-1][f._v-1] = 1;
      return true;
    }
/*
  assert(_A[f._v-1][f._w-1] == 1);
  assert(_A[f._w-1][f._v-1] == 1);
*/
  return false;
}

//----------------------------------------------------------------------------------------------

// Add every edge f in F to G; return the number of edges not previously present in G.
int CGraph::add_edges(const CEdges& F)
{
  int added = 0;
  vector<CEdge>::const_iterator f;
  for (f = F._E.begin(); f != F._E.end(); f++)
    {
      if (add_edge(*f)) added++;
    }

  return added;
}

//----------------------------------------------------------------------------------------------

// Remove edge d from G; return if d was previously present in G.
bool CGraph::remove_edge(const CEdge& d)
{
  if (_A[d._v-1][d._w-1] == 1)
    {
/*
      assert(_A[d._w-1][d._v-1] == 1);
*/
      _A[d._v-1][d._w-1] = 0;
      _A[d._w-1][d._v-1] = 0;
      return true;
    }
/*
  assert(_A[d._v-1][d._w-1] == 0);
  assert(_A[d._w-1][d._v-1] == 0);
*/
  return false;
}

//----------------------------------------------------------------------------------------------

// Remove every edge d in D from G; return the number of edges previously present in G.
int CGraph::remove_edges(const CEdges& D)
{
  int removed = 0;
  vector<CEdge>::const_iterator d;
  for (d = D._E.begin(); d != D._E.end(); d++)
    {
      if (remove_edge(*d)) removed++;
    }

  return removed;
}

//----------------------------------------------------------------------------------------------

// Report graph
void CGraph::report(void) const
{ 
  cout << "graph on " << _n << " vertices " << endl;
  cout << "with edges " << "{";
  int e = 0;
  for (int i=0; i<_n; i++)
    {
      for (int j=i; j<_n; j++)
        {
          if (_A[i][j] == 1)
            {
              if (++e > 1) cout << ", ";
              cout << "<" << i+1 << ", " << j+1 << ">";
            }
        }
    }
  cout << "}";

};

//----------------------------------------------------------------------------------------------

// Stream out graph
ostream& operator<<(ostream& os, const CGraph& G) 
{
  os << "[";
  int e = 0;
  for (int i=0; i<G._n; i++)
    {
      for (int j=i; j<G._n; j++)
        {
          if (G._A[i][j] == 1)
            {
              if (++e > 1) os << ", ";
              os << "{" << i+1 << ", " << j+1 << "}";
            }
        }
    }
  os << "]";

  return os;
}

//----------------------------------------------------------------------------------------------
