#ifndef __UTILITIES_H__
#define __UTILITIES_H__

#include "matrix.h"

//----------------------------------------------------------------------------------------------

#if (VCPP6_COMPILER == 1)

template<class T> 
inline T min(const T& a, const T& b) { return (a < b) ? a : b; };

template<class T> 
inline T max(const T& a, const T& b) { return (a > b) ? a : b; };

#endif

template<class T> T min(const vector<T>& x)
{
  int k = x.size();
  if (k == 0) ERROR("In min: Minimum of empty vector undefined. ")

  int x_min = x[0];
  for (int i = 1; i < k; i++)
    {
      if (x[i] < x_min) x_min = x[i];
    }

  return x_min;
}

template<class T> T max(const vector<T>& x)
{
  int k = x.size();
  if (k == 0) ERROR("In max: Maximum of empty vector undefined. ")

  int x_max = x[0];
  for (int i = 1; i < k; i++)
    {
      if (x[i] > x_max) x_max = x[i];
    }

  return x_max;
}

//----------------------------------------------------------------------------------------------

// Vector and matrix manipulation

// Create a vector x = (x_1, ..., x_k)
template<class T> 
vector<T>& create_vector(vector<T>& x, int k, ...)
{
  x.resize(k);

  va_list ap;
  va_start(ap, k);
  for (int i = 0; i < k; i++)
    {
      x[i] = va_arg(ap, T);
    }
  va_end(ap);

  return x;
}

// Print a vector x = (x_1, ..., x_k)
template<class T> 
ostream& operator<<(ostream& os, const vector<T>& x)
{
  int k = x.size();
  os << "(";
  for (int i = 0; i < k; i++)
    {
      if (i > 0) os << ", ";
      os << x[i];
    }
  os << ")";

  return os;
}

// Create a matrix M = ((m_11, ..., m_1k), ..., (m_k1, ..., m_kk))
template<class T> 
vector< vector<T> >& create_matrix(vector< vector<T> >& M, int k, ...)
{
  M.clear();
  M.resize(k, vector<T>(k));
  int k2 = k*k;

  va_list ap;
  va_start(ap, k);
  for (int i = 0; i < k2; i++)
    {
      M[i/k][i%k] = va_arg(ap, T);
    }
  va_end(ap);

  return M;
}

// Print a matrix M = ((m_11, ..., m_1k), ..., (m_k1, ..., m_kk))
template<class T> 
ostream& operator<<(ostream& os, const matrix<T>& M)
{
  int k = M.size();
  if (k == 0) 
    {
      os << "()";
      return os;
    }

  int l = M[0].size();
  for (int i = 0; i < k; i++)
    {
      if (M[i].size() != l) ERROR("In printmatrix: M not a square matrix. ")

      // It is assumed that a new line was started with the string "M = ",
      // hence we leave four spaces in front of every new line.
      if (i == 0) os << "/ ";
      else if (i == k-1) os << "    \\ ";
      else os << "    | ";

      for (int j = 0; j < l; j++)
        {
          if (j > 0) os << ", ";
          os << M[i][j];
        }

      if (i == 0) os << " \\" << endl;
      else if (i == k-1) os << " /";
      else os << " |" << endl;
    }

  return os;
}

//----------------------------------------------------------------------------------------------

// String operations

// Number to string conversion
template<class T> 
string n2s(T x)
{
  stringstream strx;
  strx << x;
  return strx.str();
}

#if (VCPP6_COMPILER == 1)

// T to string concatenation
template<class T>
string operator+(const string& s, T x)
{
  stringstream strsx;
  strsx << s << x;
  return strsx.str();
}

#elif (GCC_COMPILER == 1)

// int to string concatenation
inline string operator+(const string& s, int x)
{
  stringstream strsx;
  strsx << s << x;
  return strsx.str();
}

#endif

//----------------------------------------------------------------------------------------------

// Plotting

// Plot table Z with row labels X, column labels Y and title t to the specified file
template<class S, class T, class U, class V> 
void plotXYZ(const vector<S>& X, const vector<T>& Y, const matrix<U>& Z, const V& t, const string& filename)
{
  int R = X.size();
  int r;

  int C = Y.size();
  int c;

  // Validate input
  if ((R <= 0) || (C <= 0)) ERROR("In plotXYZ: empty plot. ")

  if (Z.size() != R) ERROR("In plotXYZ: table Z has wrong number of rows. ")

  if (Z[0].size() != C) ERROR("In plotXYZ: table Z has wrong number of columns. ")

  for (r = 1; r < R; r++) if (Z[r].size() != C) ERROR("In plotXYZ: table Z not an NxM square matrix. ")

  // Open output file
  ofstream os(filename.c_str());

  // Plot column labels
  os << t << "\t";
  for (c = 0; c < C; c++) os << Y[c] << "\t";
  os << endl;

  // Plot rows, with leading row label
  for (r = 0; r < R; r++)
    {
      os << X[r] << "\t";
      for (c = 0; c < C; c++) 
        {
          os << Z[r][c] << "\t";
        }
      os << endl;
    }

  os.close();
}

//----------------------------------------------------------------------------------------------

#endif
